/****************************************************************************/

/* Includes */

#include <windows.h>
#include <stdio.h>

/****************************************************************************/

/* Defines */

/* Messages passed to device notification callbacks: passed as part of
 * am_device_notification_callback_info. */

#define ADNCI_MSG_CONNECTED     1
#define ADNCI_MSG_DISCONNECTED  2
#define ADNCI_MSG_UNKNOWN       3

#define BYTES 16

//#define SNIFFER 1   // Uncomment to sniff the traffic

/****************************************************************************/

/* Estructuras */

struct am_device
{
  unsigned char unknown0[16]; /* 0 - zero */
  unsigned int device_id;     /* 16 */
  unsigned int product_id;    /* 20 - set to AMD_IPHONE_PRODUCT_ID */
  char *serial;               /* 24 - set to AMD_IPHONE_SERIAL */
  unsigned int unknown1;      /* 28 */
  unsigned char unknown2[4];  /* 32 */
  unsigned int lockdown_conn; /* 36 */
  unsigned char unknown3[8];  /* 40 */
};

struct am_device_notification_callback_info
{
  struct am_device *dev;  /* 0    device */
  unsigned int msg;       /* 4    one of ADNCI_MSG_* */
};

struct connection
{
  SOCKET client_handle;
  SOCKET iphone_handle;
};

/****************************************************************************/

/* Prototipos */

int patchload_iTunes_library ( void );
void wait_connections ( void );
void notification ( struct am_device_notification_callback_info * );
void connection_client_to_iphone ( struct connection * );
void connection_iphone_to_client ( struct connection * );
void patch_code ( void );

int PASCAL FAR my_connect ( IN SOCKET s, IN const struct sockaddr FAR *name, IN int namelen );
int PASCAL FAR my_send ( IN SOCKET s, IN const char FAR * buf, IN int len, IN int flags );
int PASCAL FAR my_recv ( IN SOCKET s, OUT char FAR * buf, IN int len, IN int flags );

int WindowsSocketInit ( void );
void imprimir_patron ( unsigned int , char * , unsigned int );
void hook_winsock_function ( HMODULE , void * , void * , void * );

/****************************************************************************/

/* Variables Globales */

int ( *AMDeviceNotificationSubscribe ) ( void * , int , int , int , char * );
int ( *AMDeviceConnect ) ( struct am_device *device );
int ( *AMDeviceIsPaired ) ( struct am_device *device );
int ( *AMDeviceValidatePairing ) ( struct am_device *device );
int ( *AMDeviceStartSession ) ( struct am_device *device );
int ( *AMDeviceStartService ) ( struct am_device *device, char *service_name, int *handle, unsigned int *unknown );
int ( *AFCConnectionOpen ) ( int handle, unsigned int io_timeout, struct afc_connection ** );
                                   
/* Libreria para poder comunicarme con el servicio */
char *iTunesLibraryName = "iTunesMobileDevice.dll";

char *service_name =  "\x00\x00\x00\x00\x8c\x07\x01\x00\x0d"
                      "com.apple.afc\x00";

/* Handle de la dll a cargar */
HMODULE iTunesHandle;

/* Variables usadas para hookear la conexion hecha hacia el iPhone */
int cantidad_conexiones = 0;
SOCKET ultimo_handle = 0;

unsigned int port;
char *string;
unsigned int local_port;

SOCKET new_sock;

int descargar_lib = FALSE;

/****************************************************************************/

/* Programa */

int main ( int argc , char *argv [] )
{
/* Controlo los parametros */
  if ( argc != 3 )
  {
    printf ( "\niphone_tunnel v1.01" );
    printf ( "\nCreated by Nicolas A. Economou ( neconomou@corest.com )" );
    printf ( "\nBuenos Aires, Argentina ( 2008 )\n" );
    printf ( "\nusage: iphone_tunnel iphone_port local_port\n" );
    printf ( "example: iphone_tunnel 22 22\n" );
    return ( 0 );
  }

/* Guardo el puerto al que me quiero conectar */
  sscanf ( argv [ 1 ] , "%i" , &port );
  port = htons ( ( unsigned short int ) port );

/* Guardo el puerto local al cual voy a esperar clientes */
  sscanf ( argv [ 2 ] , "%i" , &local_port );
  local_port = htons ( ( unsigned short int ) local_port );

/* Me pongo a esperar conexiones */
  wait_connections ();

  return ( 0 );
}

/****************************************************************************/

/* Funciones */

int patchload_iTunes_library ( void )
{
  void *ptr_connect_ws2_32;
  void *ptr_send_ws2_32;
  void *ptr_recv_ws2_32;
  void *ptr_connect_wsock32;
  void *ptr_send_wsock32;
  void *ptr_recv_wsock32;
  int ret = TRUE;

/* Cargo la dll */
  iTunesHandle = LoadLibrary ( iTunesLibraryName );

/* Si no pudo cargar la DLL */
  if ( iTunesHandle == NULL )
  {
    printf ( "iTunesMobileDevice.dll not found !\n" );
    return ( FALSE );
  }

////////////////////////////////////

/* Inicializo las variables de control para hookear la conexion */
  cantidad_conexiones = 0;
  ultimo_handle = 0;

/* Funciones a Hookear */
  ptr_connect_ws2_32  = GetProcAddress ( GetModuleHandle ( "ws2_32.dll" ) , "connect" );
  ptr_send_ws2_32     = GetProcAddress ( GetModuleHandle ( "ws2_32.dll" ) , "send" );
  ptr_recv_ws2_32     = GetProcAddress ( GetModuleHandle ( "ws2_32.dll" ) , "recv" );
  ptr_connect_wsock32 = GetProcAddress ( GetModuleHandle ( "wsock32.dll" ) , "connect" );
  ptr_send_wsock32    = GetProcAddress ( GetModuleHandle ( "wsock32.dll" ) , "send" );
  ptr_recv_wsock32    = GetProcAddress ( GetModuleHandle ( "wsock32.dll" ) , "recv" );

/* Hookeo las funciones en la IAT de iTunesMobileDevice.dll */
  hook_winsock_function ( iTunesHandle , ptr_connect_ws2_32 , ptr_connect_wsock32 , ( void * ) my_connect );
  hook_winsock_function ( iTunesHandle , ptr_send_ws2_32 , ptr_send_wsock32 , ( void * ) my_send );
  hook_winsock_function ( iTunesHandle , ptr_recv_ws2_32 , ptr_recv_wsock32 , ( void * ) my_recv );

////////////////////////////////////

/* Resuelvo los nombres */
  AMDeviceNotificationSubscribe = ( void * ) GetProcAddress ( iTunesHandle , "AMDeviceNotificationSubscribe" );
  AMDeviceConnect               = ( void * ) GetProcAddress ( iTunesHandle , "AMDeviceConnect" );
  AMDeviceIsPaired              = ( void * ) GetProcAddress ( iTunesHandle , "AMDeviceIsPaired" );
  AMDeviceValidatePairing       = ( void * ) GetProcAddress ( iTunesHandle , "AMDeviceValidatePairing" );
  AMDeviceStartSession          = ( void * ) GetProcAddress ( iTunesHandle , "AMDeviceStartSession" );
  AMDeviceStartService          = ( void * ) GetProcAddress ( iTunesHandle , "AMDeviceStartService" );
  AFCConnectionOpen             = ( void * ) GetProcAddress ( iTunesHandle , "AFCConnectionOpen" );

/* Doy permiso para poder descargar la lib */
  descargar_lib = TRUE;

  return ( ret );
}

/****************************************************************************/

void hook_winsock_function ( HMODULE iTunesHandle , void *ptr_ws2_32 , void *ptr_wsock32 , void *ptr_hook )
{
  unsigned char *ptr;
  unsigned int old;

/* Busco un ptr a la funcion en la IAT de iTunesMobileDevice.dll */
  for ( ptr = ( unsigned char * ) iTunesHandle ; ptr < ( unsigned char * ) 0xffffffff ; ptr++ )
  {

  /* Si encontre el puntero */
    if (
        ( memcmp ( ptr , &ptr_ws2_32 , sizeof ( void * ) ) == 0 ) ||
        ( memcmp ( ptr , &ptr_wsock32 , sizeof ( void * ) ) == 0 )
       )
    {
    /* Piso la IAT de iTunes de la funcion */
      VirtualProtect ( ( void * ) ptr , 4, PAGE_READWRITE , ( void * ) &old );
      * ( void ** ) ptr = ptr_hook;
      VirtualProtect ( ( void * ) ptr , 4, old , ( void * ) &old );

    /* Salgo */
      break;
    }
  }
}

/****************************************************************************/

void wait_connections ( void )
{
  SOCKET sock;
  char notif [ 1024 ];
  struct sockaddr_in saddr, saddrin;
  int len;
  int ret;

/* Inicializo el socket */
  WindowsSocketInit ();

/* Creo un socket para aceptar conexiones */
  saddr.sin_family = AF_INET;
  saddr.sin_addr.s_addr = INADDR_ANY;
  saddr.sin_port = ( unsigned short ) local_port;     
  sock = socket ( AF_INET , SOCK_STREAM , 0 );

/* Preparo la conexion */
  ret = bind ( sock , ( struct sockaddr * ) &saddr , sizeof ( struct sockaddr ) );

/* Seteo los parametros de la conexion */
  listen ( sock , 0 );

/* Si no pude bindear el puerto */
  if ( ret == SOCKET_ERROR )
  {
    printf ( "bind error !\n" );
    return;
  }

/* Si no pude cargar la DLL */
  if ( patchload_iTunes_library () == FALSE )
  {
    return;
  }

/* Espero conexiones */
  while ( ( new_sock = accept ( sock , ( struct sockaddr * ) &saddrin , &len ) ) != -1 )
  {
    printf ( "\nnew connection !\n" );

  /* Registro un callback */
    ret = AMDeviceNotificationSubscribe ( notification , 0 , 0 , 0 , notif );
    printf ( "AMDeviceNotificationSubscribe = %i\n" , ret );

  /* Espero a que el callback setee esta variable */
    while ( 1 )
    {
    /* Si la variable cambio */
      if ( descargar_lib == TRUE )
      {
      /* Deseteo la variable */
        descargar_lib = FALSE;

      /* Salgo */
        break;
      }

    /* Hago una pausa para no usar todo el CPU */
      Sleep ( 1000 );
    }

  /* Una vez que la variable fue seteada le doy tiempo al callback para que retorne */
    Sleep ( 2000 );

  /* Descargo la lib */
    FreeLibrary ( iTunesHandle ); 

  /* Si no pude cargar la DLL */
    if ( patchload_iTunes_library () == FALSE )
    {
      return;
    }

  /* Si no puedo setear el callback */
    if ( ret != 0 )
    {
      printf ( "callback error !\n" );
      return;
    }
  }
}

/****************************************************************************/

void notification ( struct am_device_notification_callback_info *info )
{
  struct afc_connection *connection;
  struct connection *conexion1;
  struct connection *conexion2;
  DWORD lpThreadId;
  SOCKET handle;
  int ret;

/* Si se pudo conectar al iPhone */
  if ( info -> msg == ADNCI_MSG_CONNECTED )
  {
    puts ( "iPhone attached !" );

  /* Me conecto al dispositivo */
    ret = AMDeviceConnect ( info -> dev );
    printf ( "AMDeviceConnect = %i\n" , ret );

    ret = AMDeviceIsPaired ( info -> dev );
    printf ( "AMDeviceIsPaired = %i\n" , ret );

    ret = AMDeviceValidatePairing ( info -> dev );
    printf ( "AMDeviceValidatePairing = %i\n" , ret );

    ret = AMDeviceStartSession ( info -> dev );
    printf ( "AMDeviceStartSession = %i\n" , ret );

  /* Levanto un servicio */
    ret = AMDeviceStartService ( info -> dev, service_name, ( int * ) &handle, NULL );
    printf ( "AMDeviceStartService = %i\n" , ret );

  /* Me conecto al servicio */
    ret = AFCConnectionOpen ( handle , 0, &connection );
    printf ( "AFCConnectionOpen = %i\n" , ret );

  /* Creo un estructura para cada thread */
    conexion1 = malloc ( sizeof ( struct connection ) );    
    conexion2 = malloc ( sizeof ( struct connection ) );    

  /* Seteo los valores de los handles de la conexiones para que lo usen los threads */
  /* Creo 2 estructuras iguales as cada thread puede eliminar la suya */
    conexion1 -> client_handle = ( SOCKET ) new_sock;
    conexion1 -> iphone_handle = ( SOCKET ) handle;
    conexion2 -> client_handle = ( SOCKET ) new_sock;
    conexion2 -> iphone_handle = ( SOCKET ) handle;

  /* Conexion iPhone -> client */
    ret = CreateThread ( NULL , 0 , ( LPTHREAD_START_ROUTINE ) connection_client_to_iphone , ( void * ) conexion1 , 0 , &lpThreadId );

  /* Conexion client -> iPhone */
    ret = CreateThread ( NULL , 0 , ( LPTHREAD_START_ROUTINE ) connection_iphone_to_client , ( void * ) conexion2 , 0 , &lpThreadId );
  }
  else
  {
    puts ( "FUIRA" );
  }

/* Descargo la DLL */
  descargar_lib = TRUE;

/* Cierro este thread */
  ExitThread ( 1 );
}

/****************************************************************************/

void connection_client_to_iphone ( struct connection *conexion )
{
  char buffer [ 256 ];
  int bytes_leidos;
  int bytes_enviados;

  while ( 1 )
  {
  /* Recibo data del cliente */
    bytes_leidos = recv ( conexion -> client_handle , buffer , 256 , 0 );

  /* Mando data al server */
    bytes_enviados = send ( conexion -> iphone_handle , buffer , bytes_leidos , 0 );

  /* Si hubo algun error */
    if ( bytes_leidos == 0 || bytes_leidos == SOCKET_ERROR || bytes_enviados == 0 || bytes_enviados == SOCKET_ERROR )
    {
    /* Cierro las 2 conexiones */
      closesocket ( conexion -> client_handle );
      closesocket ( conexion -> iphone_handle );

    /* Libero la estructura */
      free ( conexion );

    /* Cierro el thread */
      return;
    }
  }
}

/****************************************************************************/

void connection_iphone_to_client ( struct connection *conexion )
{
  char buffer [ 256 ];
  int bytes_leidos;
  int bytes_enviados;

  while ( 1 )
  {
  /* Recibo data del server */
    bytes_leidos = recv ( conexion -> iphone_handle , buffer , 256 , 0 );

  /* Mando data al cliente */
    bytes_enviados = send ( conexion -> client_handle , buffer , bytes_leidos , 0 );

  /* Si hubo algun error */
    if ( bytes_leidos == 0 || bytes_leidos == SOCKET_ERROR || bytes_enviados == 0 || bytes_enviados == SOCKET_ERROR )
    {
    /* Cierro las 2 conexiones */
      closesocket ( conexion -> client_handle );
      closesocket ( conexion -> iphone_handle );

    /* Libero la estructura */
      free ( conexion );

    /* Cierro el thread */
      return;
    }
  }
}

/****************************************************************************/

int PASCAL FAR my_connect ( IN SOCKET s, IN const struct sockaddr FAR *name, IN int namelen )
{
  int r;
  int port;
  char *host;

#ifdef SNIFFER
  if ( name -> sa_family == AF_INET )
    {
    host = (char *) inet_ntoa ( ( ( struct sockaddr_in * ) name ) -> sin_addr );
    port = ntohs ( ( ( struct sockaddr_in * ) name ) -> sin_port );
    printf ( "connect: socket %i, host %s port %i ... " , s , host , port );
#endif

    r = connect( s , name , namelen );

#ifdef SNIFFER
    if( r == 0 )
    {
      printf ( "connected\n\n" );
    } else {
      printf ( "NOT connected\n\n" );
    }
    fflush ( stdout );
  }
#endif

  return ( r );
}

/****************************************************************************/

int PASCAL FAR my_send ( IN SOCKET s, IN const char FAR * buf, IN int len, IN int flags)
{
  int r;

/* Si el handle no es igual al anterior */
  if ( s != ultimo_handle )
  {
  /* Tengo una nueva conexion */
    ultimo_handle = s;

  /* Incremento la cantidad de conexiones */
    cantidad_conexiones ++;
  }

/* Si es el paquete donde va el puerto donde esta escuchando el afcd */
  if ( ( len == 0x18 ) && ( cantidad_conexiones == 3 ) )
  {
  /* Cambio el puerto al paquete */
    memcpy ( ( void * ) &buf [ 0x14 ] , ( void * ) &port , 2 );

  /* Reseteo las variables para poder realizar otra conexion */
    cantidad_conexiones = 0;  
    ultimo_handle = 0;
  }

  r = send ( s, buf, len, flags );

#ifdef SNIFFER
  printf ( "send: socket %i, %i bytes\n" , s , r );
  if ( r > 0 )
  {
    imprimir_patron ( 0 , ( char * ) buf , r );
  }
  printf ( "\n\n" );
  fflush ( stdout );
#endif
  
  return ( r );
}

/****************************************************************************/

int PASCAL FAR my_recv ( IN SOCKET s, OUT char FAR * buf, IN int len, IN int flags)
{
  int r;

  r = recv ( s, buf, len, flags );

#ifdef SNIFFER
  printf ( "recv: socket %i, %i bytes\n" , s , r );
  if ( r > 0 )
  {
    imprimir_patron ( 0 , ( char * ) buf , r );
  }
  printf ( "\n\n" );
  fflush ( stdout );
#endif

  return ( r );
}

/****************************************************************************/

void imprimir_patron ( unsigned int direccion , char *buffer , unsigned int bytes_a_imprimir )
{
  unsigned int cont;
  unsigned int i;

/* Imprimo las lineas encontradas */
  for ( cont = 0 ; cont < bytes_a_imprimir ; cont = cont + BYTES )
  {
  /* Imprimo la direccion de la memoria */
    printf ( "%.8x | " , direccion );

  /* Incremento la direccion a mostrar */
    direccion = direccion + BYTES;
    
  /* Imprimo en hexa */
    for ( i = 0 ; i < BYTES ; i ++ )
    {
    /* Imprimo la cantidad que pedi */  
      if ( i < ( bytes_a_imprimir - cont ) )
      {
        printf ( "%.2x " , ( unsigned char ) buffer [ i + cont ] );
      }
      else
      {
        printf ( "   " );
      }
    }

  /* Espacio entre las 2 columnas */
    printf ( "| " );

  /* Imprimo en caracteres */  
    for ( i = 0 ; i < BYTES ; i ++ )
    {
      if ( i < ( bytes_a_imprimir - cont ) )
      {
        printf ( "%c" , ( isgraph ( buffer [ i + cont ] ) ) ? buffer [ i + cont ] : '.' );
      }
      else
      {
        printf ( " " );
      }
    }

  /* Fin de linea */
    printf ( "\n" );
  }
}

/****************************************************************************/

int WindowsSocketInit ( void )
{
  static int inicializado = 0;
  WORD wRequestedVersion;
  WSADATA wsaData;
  int err;

/* Si no fue inicializado */
  if ( inicializado == FALSE )
  {
/* Versin 2.2 de los sockets */
    wRequestedVersion = 0x0202;
    err = WSAStartup ( wRequestedVersion , &wsaData );

    if ( err == 0 )
    {
    /* Si la versin que soporta la WinSock DLL del sistema no es suficiente */
    /* Indicar el error al usuario, limpiar y salir */

      if ( wsaData.wVersion != wRequestedVersion )
      {
        WSACleanup();
      }
      else
      {
        inicializado = TRUE;
      }
    }
  }

  return ( inicializado );
}

/****************************************************************************/
